@@ -4,9 +4,8 @@ import utils
 import analyze
 import step_methods as steps
 import generate
-import distributions
 
-__version__ = '0.5beta'
+__version__ = '0.5.2.beta.dev'
 
 try:
     from IPython.core.debugger import Tracer; debug_here = Tracer()
