@@ -1,5 +1,6 @@
 from __future__ import division
 import sys, os
+from types import FunctionType
 
 import numpy as np
 from matplotlib.pylab import figure
@@ -215,7 +216,7 @@ def group_cond_diff(hm, node, cond1, cond2, threshold=0):
 
     return pooled_mean, pooled_var, mass_under
 
-def _evaluate_post_pred(sampled_stats, data_stats, evals=None):
+def post_pred_compare_stats(sampled_stats, data_stats, evals=None):
     """Evaluate summary statistics of sampled sets.
 
     :Arguments:
@@ -234,23 +235,26 @@ def _evaluate_post_pred(sampled_stats, data_stats, evals=None):
         # Generate some default evals
         evals = OrderedDict()
         evals['observed'] = lambda x, y: y
-        evals['credible'] = lambda x, y: (scoreatpercentile(x, 97.5) > y) and (scoreatpercentile(x, 2.5) < y)
-        evals['quantile'] = percentileofscore
-        evals['SEM'] = lambda x, y: (np.mean(x) - y)**2
-        evals['mahalanobis'] = lambda x, y: np.abs(np.mean(x) - y)/np.std(x)
         evals['mean'] = lambda x,y: np.mean(x)
         evals['std'] = lambda x,y: np.std(x)
-        for q in [2.5, 25, 50, 75, 97.5]:
-            key = str(q) + 'q'
-            evals[key] = lambda x, y, q=q: scoreatpercentile(x, q)
+        evals['SEM'] = lambda x, y: (np.mean(x) - y)**2
+        evals['MSE'] = lambda x, y: np.mean((x - y)**2)
+        evals['credible'] = lambda x, y: (scoreatpercentile(x, 97.5) > y) and (scoreatpercentile(x, 2.5) < y)
+        evals['quantile'] = percentileofscore
+        evals['mahalanobis'] = lambda x, y: np.abs(np.mean(x) - y)/np.std(x)
+        #for q in [2.5, 25, 50, 75, 97.5]:
+        #    key = str(q) + 'q'
+        #    evals[key] = lambda x, y, q=q: scoreatpercentile(x, q)
 
     # Evaluate all eval-functions
-    results = pd.DataFrame(index=sampled_stats.keys(), columns=evals.keys() + ['NaN'])
+    results = pd.DataFrame(index=sampled_stats.keys(), columns=evals.keys() + ['NaN'],
+                           dtype=np.float32)
+
     results.index.names = ['stat']
-    for stat_name in sampled_stats.iterkeys():
+    for stat_name in sampled_stats:
         #update NaN column with the no. of NaNs and remove them
         s = sampled_stats[stat_name]
-        results.ix[stat_name]['NaN'] = sum(np.isnan(s))
+        results.ix[stat_name]['NaN'] = sum(pd.isnull(s))
         s = s[np.isfinite(s)]
         if len(s) == 0:
             continue
@@ -259,53 +263,26 @@ def _evaluate_post_pred(sampled_stats, data_stats, evals=None):
             value = func(s, data_stats[stat_name])
             results.ix[stat_name][eval_name] = value
 
-    return results
+    return results.drop('NaN', axis=1)
 
 
-def _post_pred_summary_bottom_node(bottom_node, samples=500, stats=None, plot=False, bins=100, evals=None):
-    """Create posterior predictive check for a single bottom node."""
-    def _calc_stats(data, stats):
-        out = {}
-        for name, func in stats.iteritems():
-            out[name] = func(data)
-        return out
-
-    if stats is None:
-        stats = OrderedDict((('mean', np.mean), ('std', np.std)))
-
-    ############################
-    # Compute stats over data
-    data = bottom_node.value
-    data_stats = _calc_stats(data, stats)
-
-    ###############################################
-    # Initialize posterior sample stats container
-    sampled_stats = {}
-    for name in stats.iterkeys():
-        sampled_stats[name] = np.empty(samples)
+def _post_pred_generate(bottom_node, samples=500, data=None, append_data=False):
+    """Generate posterior predictive data from a single observed node."""
+    datasets = []
 
     ##############################
     # Sample and generate stats
     for sample in range(samples):
         _parents_to_random_posterior_sample(bottom_node)
         # Generate data from bottom node
-        sampled = bottom_node.random()
-        sampled_stat = _calc_stats(sampled, stats)
+        sampled_data = bottom_node.random()
+        if append_data and data is not None:
+            sampled_data = sampled_data.join(data, lsuffix='_sampled')
+        datasets.append(sampled_data)
 
-        # Add it the results container
-        for name, value in sampled_stat.iteritems():
-            sampled_stats[name][sample] = value
+    return datasets
 
-    if plot:
-        from pymc.Matplot import gof_plot
-        for name, value in sampled_stats.iteritems():
-            gof_plot(value, data_stats[name], nbins=bins, name=name, verbose=0)
-
-    result = _evaluate_post_pred(sampled_stats, data_stats, evals=evals)
-
-    return result
-
-def post_pred_check(model, samples=500, bins=100, stats=None, evals=None, plot=False, progress_bar=False):
+def post_pred_gen(model, groupby=None, samples=500, append_data=False, progress_bar=True):
     """Run posterior predictive check on a model.
 
     :Arguments:
@@ -315,12 +292,76 @@ def post_pred_check(model, samples=500, bins=100, stats=None, evals=None, plot=F
     :Optional:
         samples : int
             How many samples to generate for each node.
+        groupby : list
+            Alternative grouping of the data. If not supplied, uses splitting
+            of the model (as provided by depends_on).
+        append_data : bool (default=False)
+            Whether to append the observed data of each node to the replicatons.
+        progress_bar : bool (default=True)
+            Display progress bar
+
+    :Returns:
+        Hierarchical pandas.DataFrame with multiple sampled RT data sets.
+        1st level: wfpt node
+        2nd level: posterior predictive sample
+        3rd level: original data index
+
+    :See also:
+        post_pred_stats
+    """
+    results = {}
+
+    # Progress bar
+    if progress_bar:
+        n_iter = len(model.get_observeds())
+        bar = pbar.ProgressBar(n_iter)
+        bar_iter = 0
+    else:
+        print "Sampling..."
+
+    if groupby is None:
+        iter_data = ((name, model.data.ix[obs['node'].value.index]) for name, obs in model.iter_observeds())
+    else:
+        iter_data = model.data.groupby(groupby)
+
+    for name, data in iter_data:
+        node = model.get_data_nodes(data.index)
+
+        if progress_bar:
+            bar_iter += 1
+            bar.animate(bar_iter)
+
+        if node is None or not hasattr(node, 'random'):
+            continue # Skip
+
+        ##############################
+        # Sample and generate stats
+        datasets = _post_pred_generate(node, samples=samples, data=data, append_data=append_data)
+        results[name] = pd.concat(datasets, names=['sample'], keys=range(len(datasets)))
+
+        if progress_bar:
+            bar.animate(n_iter)
+
+    return pd.concat(results, names=['node'])
+
+
+def post_pred_stats(data, sim_datasets, stats=None, plot=False, bins=100, evals=None, call_compare=True):
+    """Calculate a set of summary statistics over posterior predictives.
+
+    :Arguments:
+        data : pandas.Series
+
+        sim_data : pandas.Series
+
+    :Optional:
         bins : int
             How many bins to use for computing the histogram.
-        stats : dict
+        stats : dict or function
             User-defined statistics to compute (by default mean and std are computed)
             and evaluate over the samples.
-            :Example: {'mean': np.mean, 'median': np.median}
+            :Example:
+              * {'mean': np.mean, 'median': np.median}
+              * lambda x: np.mean(x)
         evals : dict
             User-defined evaluations of the statistics (by default 95 percentile and SEM).
             :Example: {'percentile': scoreatpercentile}
@@ -328,38 +369,50 @@ def post_pred_check(model, samples=500, bins=100, stats=None, evals=None, plot=F
             Whether to plot the posterior predictive distributions.
         progress_bar : bool
             Display progress bar while sampling.
-
-
-    :Returns:
-        Hierarchical pandas.DataFrame with the different statistics.
-
+        field : string
+            Which column name to run the stats on
+        call_com,pare : bool (default=True)
+            Whether to call post_pred_compare_stats. If False, return stats directly.
     """
-    import pandas as pd
-    results = {}
 
-    # Progress bar
-    if progress_bar:
-        n_iter = len(model.get_observeds()) * model.num_subjs
-        bar = pbar.ProgressBar(n_iter)
-        bar_iter = 0
+    def _calc_stats(data, stats):
+        out = {}
+        for name, func in stats.iteritems():
+            out[name] = func(data)
+        return out
+
+    if stats is None:
+        stats = OrderedDict((('mean', np.mean), ('std', np.std)))
+    if isinstance(stats, FunctionType):
+        stats = OrderedDict((('stat', stats),))
+
+    data_stats = _calc_stats(data, stats)
+
+    ###############################################
+    # Initialize posterior sample stats container
+    samples = len(sim_datasets)
+    sampled_stats = {}
+    sampled_stats = pd.DataFrame(index=sim_datasets.index.droplevel(2).unique(),
+                                 columns=stats.keys(),
+                                 dtype=np.float32)
+
+    for i, sim_dataset in sim_datasets.groupby(level=(0, 1)):
+        sampled_stat = _calc_stats(sim_dataset.values, stats)
+
+        # Add it to the results container
+        for name, value in sampled_stat.iteritems():
+            sampled_stats[name][i] = value
+
+    if plot:
+        from pymc.Matplot import gof_plot
+        for name, value in sampled_stats.iteritems():
+            gof_plot(value, data_stats[name], nbins=bins, name=name, verbose=0)
+
+    if call_compare:
+        return post_pred_compare_stats(sampled_stats, data_stats, evals=evals)
     else:
-        print "Sampling..."
+        return sampled_stats
 
-    for name, obs_descr in model.iter_observeds():
-        node = obs_descr['node']
-
-        if progress_bar:
-            bar_iter += 1
-            bar.update(bar_iter)
-
-        if node is None or not hasattr(node, 'random'):
-            continue # Skip
-
-        results[name] = _post_pred_summary_bottom_node(node, samples=samples, bins=bins, evals=evals, stats=stats, plot=plot)
-        if progress_bar:
-            bar.animate(n_iter)
-
-    return pd.concat(results, names=['node'])
 
 def _parents_to_random_posterior_sample(bottom_node, pos=None):
     """Walks through parents and sets them to pos sample."""
