@@ -16,7 +16,8 @@ from kabuki.utils import flatten
 
 
 class Knode(object):
-    def __init__(self, pymc_node, name, depends=(), col_name='', subj=False, hidden=False, **kwargs):
+    def __init__(self, pymc_node, name, depends=(), col_name='',
+                 subj=False, hidden=False, pass_dataframe=True, **kwargs):
         self.pymc_node = pymc_node
         self.name = name
         self.kwargs = kwargs
@@ -28,6 +29,8 @@ class Knode(object):
         self.nodes = OrderedDict()
         self.hidden = hidden
 
+        self.pass_dataframe = pass_dataframe
+
         #create self.parents
         self.parents = {}
         for (name, value) in self.kwargs.iteritems():
@@ -116,7 +119,10 @@ class Knode(object):
 
             #get value for observed node
             if self.observed:
-                kwargs['value'] = grouped_data[self.col_name] #.to_records(index=False)
+                if self.pass_dataframe:
+                    kwargs['value'] = grouped_data[self.col_name] #.to_records(index=False)
+                else:
+                    kwargs['value'] = grouped_data[self.col_name].values #.to_records(index=False)
 
             # Deterministic nodes require a parent argument that is a
             # dict mapping parent names to parent nodes. Knode wraps
@@ -131,6 +137,9 @@ class Knode(object):
                     kwargs.pop(name)
                 kwargs['parents'] = parents_dict
 
+                if self.observed:
+                    kwargs['parents']['value'] = kwargs['value']
+
 
             # Deterministic nodes require a doc kwarg, we don't really
             # need that so if its not supplied, just use the name
@@ -355,6 +364,7 @@ class Hierarchical(object):
         self._setup_model()
         self.create_model()
         self.load_db(d['dbname'], db=d['db'])
+        self.gen_stats()
 
     def save(self, fname):
         """Save model to file.
@@ -739,9 +749,39 @@ class Hierarchical(object):
         for node in nodes.iterrows():
             yield node
 
+    def get_group_traces(self):
+        """Returns a DataFrame containing traces of all stochastic
+        group nodes in the model.
+        """
+        return pd.DataFrame({i.__name__: i.trace() for i in self.get_group_nodes().node})
+
+    def get_traces(self):
+        """Returns a DataFrame containing traces of all stochastic
+        nodes in the model.
+
+        :Note: It is quite easy to then save this trace to csv by
+        calling model.get_traces().to_csv('samples.csv')
+        """
+        return pd.DataFrame({i.__name__: i.trace() for i in self.get_stochastics().node})
+
+    def get_data_nodes(self, idx):
+        data_nodes = []
+        for name, node_descr in self.iter_observeds():
+            node = node_descr['node']
+            if set(idx).issubset(set(node.value.index)):
+                data_nodes.append(node)
+
+        if len(data_nodes) != 1:
+            raise NotImplementedError("Supply a grouping so that at most 1 observed node codes for each group.")
+
+        return data_nodes[0]
+
+    def __getitem__(self, name):
+        return self.nodes_db.ix[name]['node']
+
     @property
     def values(self):
-        return {name: node['node'].value[()] for (name, node) in self.iter_non_observeds()}
+        return {name: node['node'].value[()] for (name, node) in self.iter_non_observeds() if node['node'].value.shape == ()}
 
     def set_values(self, new_values):
         """
@@ -768,7 +808,7 @@ class Hierarchical(object):
             nodes : iterable
                 list nodes to optimize.
         """
-        non_observeds = filter(lambda x: x.observed, optimize_nodes)
+        non_observeds = filter(lambda x: not x.observed, optimize_nodes)
 
         init_vals = [node.value for node in non_observeds]
 
